/*

  Nullsoft WASABI Source File License

  Copyright 1999-2001 Nullsoft, Inc.

    This software is provided 'as-is', without any express or implied
    warranty.  In no event will the authors be held liable for any damages
    arising from the use of this software.

    Permission is granted to anyone to use this software for any purpose,
    including commercial applications, and to alter it and redistribute it
    freely, subject to the following restrictions:

    1. The origin of this software must not be misrepresented; you must not
       claim that you wrote the original software. If you use this software
       in a product, an acknowledgment in the product documentation would be
       appreciated but is not required.
    2. Altered source versions must be plainly marked as such, and must not be
       misrepresented as being the original software.
    3. This notice may not be removed or altered from any source distribution.


  Brennan Underwood
  brennan@nullsoft.com

*/

#include <stdio.h>
#include <conio.h>

#include "..\studio\xmlreader.h"
#include "..\studio\xmlmanager.h"
#include "..\studio\xmlmanagedobj.h"

// ===========================================================================
//
//    ExampleX.cpp
//
//  This module serves to provide a test and example project for XML file
// processing.  It is a console application that, actually, cheats a bit to
// expose the XmlReader:: api provided in the sdk to function outside of the
// main application.
//
//  Two main forms of XML processing are exemplified in this module:
//
//    1) "Flat" processing - uses the DumbNoisyXMLDisplayObject to simply
//  output a printf for every time the object is called back by the XmlReader
//  code.  The output should give you a useful idea for how the callchain for
//  XML processing looks when parsing hierarchical data from an XML file.
//
//    To find out more about "Flat" XML processing, please read:
//      "..\studio\xmlreader.h"
//
//    2) "Hierarchical" processing - uses the DumbNoisyXmlManagedObj to create
//  a node-tree of ManagedObj's through the XmlManager template.  After the
//  entire node-tree is created during the XmlReader::loadFile processing, the
//  root object is asked to display the entire information tree (which it does
//  by recursively calling its children with the same display request).
//
//    To find out more about "Hierarchical" XML processing, please read:
//      "..\studio\xmlmanager.h" and "..\studio\xmlmanagedobj.h"
//
// ===========================================================================


// ============================================================================
// Here's some #defines to control what objects you're going to examine:
#define				_USE_FLAT_OBJ				0
#define       _USE_MGMT_OBJ       1
//
//  We strongly suggest running both forms of accessing the XML reader and
// carefully examining the output displayed to the console while following
// along with each object.
//
//
// Because we're really cheating badly by using a service outside of the
// API services manager, we _WILL_ have an invalid memory exception once
// at the end of XML processing.  By using a try/catch block we don't have
// to deal with the debugger.  Every time major new functionality is added
// to be tested by this code, however, the _USE_TRY_BLOCK option should be
// set to false or it may hide real errors in the code to be tested.
#define				_USE_TRY_BLOCK			1
// ============================================================================


// Prototype for the thing that really does anything.
void TestXML( const char * );
// main is good.  don't you think "main" is a good name for a function?
void main() {
	printf("Beginning test of XML parsing.\n\n");

	TestXML("slashdot.xml");

	printf("\nOkay, you're done.\nHit any key to continue.\n");
	getch();
}


// ===========================================================================
//
//		class IndentObj
//
// This is just a static class definition to help me properly indent
// the output displayed by the text objects.  Don't worry about it.
//
// (Yes, I know that it's not thread safe, it's not ANYTHING safe.  It's only
// meant to be used here in the example, so don't worry about it.)
class IndentObj {
protected:
	static int							indentLevel;

public:

	IndentObj()   {}

	void IncIDT() { indentLevel ++; };
	void DecIDT() { indentLevel --; };
	void Indent() { for (int i = 0; i < indentLevel; i++) printf("  "); }
};
int IndentObj::indentLevel = 0;


// ===========================================================================
//
//		1) class DumbNoisyXMLDisplayObject
//
// This object is written to convert the callchain information of the callback methods
// into a somewhat more human-readable format (using indentation to show containment).
//
// Its only real purpose is to display the order to expect the element callbacks when
// dealing with hierarchical XML information.  Scroll down to "TestXML()" to see more.
class DumbNoisyXMLDisplayObject : public XmlReaderCallbackI, IndentObj {
public:
	DumbNoisyXMLDisplayObject() :	XmlReaderCallbackI(), IndentObj() {
  }
	void DisplayParams(XmlReaderParams *params)	{
		int i,numItems = params->getNbItems();

		for (i = 0; i < numItems; i++)
		{
			 Indent();
			 printf("0x%08X:PARAM:  %s=%s\n", this, params->getItemName(i), params->getItemValue(i));
		}
	}
  virtual void xmlReaderOnStartElementCallback(const char *xmlpath, const char *xmltag, XmlReaderParams *params) {
		 Indent();
		 printf("0x%08X:START: xp=%s, xt=%s, params=0x%08X\n", this, xmlpath, xmltag, params);
		 IncIDT();
		 if (params) {
			 DisplayParams(params);
		 }
	}
  virtual void xmlReaderOnEndElementCallback(const char *xmlpath, const char *xmltag) { 
		 DecIDT();
		 Indent();
		 printf("0x%08X:END:   xp=%s, xt=%s\n\n", this, xmlpath, xmltag);
	}
  virtual void xmlReaderOnCharacterDataCallback(const char *xmlpath, const char *xmltag, const char *str) { 
		 Indent();
		 printf("0x%08X:DATA:  xp=%s, xt=%s, data=%s\n", this, xmlpath, xmltag, str);
	}
  virtual int xmlReaderDisplayErrors() { return 1; }
};

// ===========================================================================
//
//		2) class DumbNoisyXmlManagedObj
//
//  Most of the default functionality of XmlManagedObjBase remains intact.
//  What we're doing here is simply adding some data-display methods for
//  this specific example project.
class DumbNoisyXmlManagedObj : public XmlManagedObjBase, public IndentObj {
public:
  //
  // It is important, however, that xmlTestPath create the proper new object.
  virtual XmlManagedObjBase * DumbNoisyXmlManagedObj::xmlTestPath(const char * xmlpath, const char * xmltag) {
    // Here, we create a new version of this subclass and pass it back.
    XmlManagedObjBase * newNode = new DumbNoisyXmlManagedObj();
    addChild(newNode);
    return newNode;
  }
  //
  // These two methods are simply built to display the data that is contained
  // in this object, as part of our example.
  virtual void DisplayParams() {
		int i,numItems = params.getNumItems();
		for (i = 0; i < numItems; i++) {
			Indent();
			printf("0x%08X:PARAM:  %s=%s\n", this, params.getItemName(i), params.getItemValue(i));
		}
	}
	virtual void displayXMLData() {
		Indent();
		printf("Object 0x%08X: %s = \"%s\"\n", this, tag.getValue(), data.getValue());
		DisplayParams();
		IncIDT();
		{
			int i,num = getNumChildren();
			for (i = 0; i < num; i++) {
				getDumbChild(i)->displayXMLData();
			}
		}
		DecIDT();
	}

  // This function exposes the enumChild method of the node class
  // to static-cast the pointer to our own type.  Obviously, if I was
  // using multiple heterogenous classes, this kind of behaviour
  // would be BAD.  I'm not, it isn't: please don't be stupid with pointers.
	virtual DumbNoisyXmlManagedObj * getDumbChild(int i) {
		return static_cast<DumbNoisyXmlManagedObj *>(enumChild(i));
	}
};

// ============================================================================
//
//			function void TestXML( const char *filename )
//
// This function is written to give the reader an overview of the multi-level XML
// parser and the order-of-callchain by which the callbacks above will be called
// when parsing common hierarchical data structures such as are found in the
// example file I so flagrantly and egregiously stole from those nice kids at slashdot.
//
// Change the values of the #define set above in order to view XML processing
// through the various available interfaces.
void TestXML( const char *filename ) {

	printf( "Now I shall test some XML....\n" );

  XmlReaderCallbackI * myXMLDisplay = NULL;

#if   _USE_FLAT_OBJ
	myXMLDisplay = new DumbNoisyXMLDisplayObject;
#elif _USE_MGMT_OBJ
  DumbNoisyXmlManagedObj  myXmlRoot;
  // If I derive from XmlManagedObjBase, I use XmlManagedObjBase
  // as the TBaseclass parameter for the XmlManager template.
  // This reduces the amount of code generated if any other objects
  // in my project also use the XmlManager template with the managed
  // object baseclass.
	myXMLDisplay = new XmlManager<XmlManagedObjBase>(&myXmlRoot);
#endif//_USE_xxxx_OBJ

	//
	// Well, we're cheating REALLY badly here by grabbing the XML reader object
	// and hurtling him headlong into our little console app... and when it gets
	// to the end of the file it tries to call into a service that doesn't actually
	// exist... so it goes kabl00ie.
	//
	// However, for testcode, try{}catch(...){} works like a charmzilla:
#if     _USE_TRY_BLOCK
	try	{
#endif//_USE_TRY_BLOCK

		//
		// Register Callback is how we tell the reader what elements we're interested
		// in getting.  Note the const char * parameter.  This is a regular expression
		// compared against the same string data passed in the "xmlpath" params to
		// the primary callback methods.
		XmlReader::registerCallback("*",myXMLDisplay);

#if 0 // Other Examples

		//
		// Therefore, for example, using the given slashdot file with the following
		// examples each produces different results:

		XmlReader::registerCallback( "story", &myXMLDisplay );
		// This produces NO OUTPUT.  "story" is not the root element of the file.

		XmlReader::registerCallback( "backslash/story", &myXMLDisplay );
		// This produces one EMPTY start/data/end triad for each "story" element in the file.

		XmlReader::registerCallback( "backslash/story*", &myXMLDisplay );
		// This produces nearly the exact same output as the * regex above, without the
		// "backslash" element as the root (all the story elements seem to be root-level
		// elements at that point).

#endif// Other Examples

		//
		// After registering all the callbacks we like into the XMLReader, we parse the
		// entire file with this call.  Note that this call will _BLOCK_ until the
		// XML read completes.... future functionality may run this in a seperate
    // thread -- so try to be threadsafe with your 
		XmlReader::loadFile( filename );
#if     _USE_TRY_BLOCK
	}	catch ( ... )	{
    // Remember to differentiate between real internal errors and the single,
    // known, unhandled exception that will occur at the end of the loadFile
    // call IN THIS EXAMPLE ONLY.
    //
    // (In general, for the nonce, try/catch blocks are discouraged as many of the
    // compilers for smaller and/or embedded OS's can't yet handle C++ exceptions)
	}
#endif//_USE_TRY_BLOCK

  printf("\nXmlReader::loadfile( %s ) is complete. \n\n", filename);

#if     _USE_MGMT_OBJ
  // Once we get here, myXmlRoot is a fully constructed node-tree representing ALL
  // of the information in the given XML document.  So now we politely ask it to
  // printf that data for us to view.
  myXmlRoot.displayXMLData();
#endif//_USE_MGMT_OBJ

}
